#include <iostream>
#include <string>
#include <thread>
#include <mutex>
#include <deque>
#include <memory>
#include <condition_variable>
#include <cstdlib>

using namespace std;

class Task
{
public:
   virtual void run() = 0;
};

typedef Task* TaskPtr;

class FakeTask : public Task
{
    string name;
public:
    FakeTask(string s) : name(s) {}

    virtual void run() override
    {
        cerr << "task name: " << name << endl;
        //do something here ...
        this_thread::sleep_for(chrono::milliseconds(50));
    }
};

//queue is implemented as monitor object 
class Queue
{
    deque<TaskPtr> list;

    //synchronization
    mutex mtx;
    condition_variable not_empty;

public:
    void addToEnd(TaskPtr t)
    {
        mtx.lock();
        list.push_back(t);
        mtx.unlock();
        not_empty.notify_one();
    }

    TaskPtr removeFirst()
    {
        unique_lock<mutex> lck(mtx);
        not_empty.wait(lck, [&]{ return list.size() > 0; });
        TaskPtr t = list[0];
        list.pop_front();
        return t;
    }
};

class ThreadPool
{
    int threadsInPool;
public:
    ThreadPool(int numberOfThreads) : threadsInPool(numberOfThreads)
    {}

    void Execute(Queue & taskQueue, Queue & completedTasks)
    {
       for (int i=0; i<threadsInPool; ++i)
       {
           thread([&]()
           {
              cerr << "running pool thread ..." << endl;

              while(1)
              {
                TaskPtr t = taskQueue.removeFirst();
                t->run();
                completedTasks.addToEnd(t);
              }

           }).detach();
       }
    }
};

int main()
{
    Queue taskQueue, completedTasks;
    FakeTask t1("name1");
    FakeTask t2("name2");

    taskQueue.addToEnd(&t1);
    taskQueue.addToEnd(&t2);

    ThreadPool pool(5);
    pool.Execute(taskQueue, completedTasks);

    //hit Enter to finish
    getchar();

    return 0;
}